#!/usr/bin/perl

#Version 1.00

use strict;

use HTTP::Proxy;
use HTTP::Proxy::BodyFilter::complete;
use HTML::TreeBuilder;

# initialisation
my $proxy = HTTP::Proxy->new( port => 3128 );
close STDERR;

# This logs the queries and clicks
$proxy->push_filter( request => LogClicks->new() );

# By forcing the client not to cache, we get all clicks, not just first-time clicks
# Depending on the setting, we might not want to do this.
$proxy->push_filter( response => NoCache->new() );

# Changes the user agent to avoid content from Bing being compressed
$proxy->push_filter( host => 'bing.com',  request => NoCompression->new());

# This catches the results, rewriting the content as well (only on Bing requests)
$proxy->push_filter( host => 'bing.com', 
					 response => HTTP::Proxy::BodyFilter::complete->new(), 
					 response => RewriteContent->new() );

$proxy->start;

############### CLICK LOGGING ################
{
    package LogClicks;
    use base qw( HTTP::Proxy::HeaderFilter );

    sub filter {
        my ( $self, $headers, $message ) = @_;

        my $url = $message->url;
		my $referer = $headers->header('Referer');
		
		# If the referer is Bing, its a follow on request
		# Only keep non-bing requests
		if ($referer =~ /bing.com\/search\?q=([^&]*)/ && $url !~ /bing.com/ && $url !~ /bing.net/) 
		{
			my $query = $1;
			my $date = scalar localtime;
			my $ip = $headers->header('X-Forwarded-For');
						
			print "CLICK\t$date\t$query\t$ip\t$url\n";
		}
    }
}

########### RESULT LOGGING, CONTENT REWRITING ############
{
    package RewriteContent;
    use base qw( HTTP::Proxy::BodyFilter );

    sub filter {
        my ( $self, $dataref, $response, $protocol, $buffer ) = @_;

        my $url = $response->base;

        # If its a Bing request, and its the last chunk of the request (all the data is present)
        if ($url =~ /^http:\/\/www.bing.com\/search?/ && !defined($buffer)) 
		{
			# Rerank and interleave results
			# 1. Cut out results, and large graphical elements (sb_ans)
			my $tree = HTML::TreeBuilder->new;
			$tree->parse_content($$dataref);
			my @results;
			foreach my $r ($tree->look_down('_tag', 'li')) 
			{
				# Standard URLs
				if ($r && $r->attr('class') eq "sa_wr") {
					$results[$#results+1] = $r->as_HTML();
				}
				# Large graphical elements
				if ($r && $r->attr('class') eq "sb_ans") {
					$results[$#results+1] = $r->as_HTML();
				}
			}
			
			# 2. Make a reranked result set
			my @reranked = (sort simpleRanker @results);
			
			# 3. Paste in a new top 10
			my $newResults = join("</li>", (@reranked)) . "</li>";
			$$dataref =~ s|(<ul[^>]*sb_results[^>]*>)<li.*?</div></li></ul>|${1}RESULTSHERE</ul>|s;
			substr($$dataref, index($$dataref, "RESULTSHERE"), length("RESULTSHERE"), $newResults);
						
			# 5. Log the query
			if ($url =~ /\?q=([^&]*)/) 
			{
				my $query = $1;
				my $date = scalar localtime;
				my $ip = $response->request->header('X-Forwarded-For');
								
				# Collect the results
				my $results = "";
				while ($$dataref =~ /<h3><a[^>]*href="(http[^"]*)"[^>]*SERP/g)
				{
					$results .= "$1;";
				}

				my $orig = record(\@results);
				print "QUERY\t$date\t$query\t$ip\t$results\t$orig\n";
			}
        }
    }
		
	# Make a list of URLs
	sub record
	{
		my $resultsRef = shift;
		my $results = join " ", @$resultsRef;

		my $output;
		while ($results =~ /<h3><a[^>]*href="([^"]*)"[^>]*SERP/g)
		{
			$output .= "$1;";
		}
		
		return $output;
	}
	
	##################### SIMPLE RERANKER #######################
	
	# Rank by how many of our favorite or least-favorite words occur in search results
	sub simpleRanker
	{
		my @goodWords = ("acm", "dblp", "edu", "2011");
		my @badWords = ("wikipedia", "answers");
		
		# $a and $b are the two input URLs. 
		my $scoreA = countWords($a, \@goodWords) - countWords($a, \@badWords);
		my $scoreB = countWords($b, \@goodWords) - countWords($b, \@badWords);
		
		# Returns +1 if URL $b scores higher, -1 if $a scores higher
		return ($scoreB <=> $scoreA);
	}
	
	# Count how many times each word occurs
	sub countWords
	{
		my ($text, $wordsRef) = (shift, shift);
		my $n = 0;
		
		foreach my $word (@$wordsRef) 
		{
			while ($text =~ /$word/ig)
			{
				$n++;
			}
		}
		return $n;
	}

}

############## HEADER REWRITING ##############
{
    package NoCompression;
    use base qw( HTTP::Proxy::HeaderFilter );

    # changes the User-Agent header in all requests
    sub filter {
        my ( $self, $headers, $message ) = @_;

		$message->headers->header( User_Agent => 'Mozilla/4.0 (compatible) Perl proxy' );
		$message->headers->header( Accept => 'application/xml,text/xml' );
    }
}

{
	package NoCache;
    use base qw( HTTP::Proxy::HeaderFilter );

	# Tells the web browser not to cache pages where the referer is Bing
    sub filter {
        my ( $self, $headers, $message ) = @_;

		my $referer = $message->request->headers->header("referer");
		if ($referer =~ /bing/) {
			$message->headers->header( 'Cache-Control' => 'no-cache' );
		}
    }
}
